"""This package implements various Nwchem Jobs and Error Handlers."""

__author__ = "Shyue Ping Ong"
__copyright__ = "Copyright 2012, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Shyue Ping Ong"
__email__ = "ongsp@ucsd.edu"
__status__ = "Beta"
__date__ = "6/17/13"



"""This module implements basic kinds of jobs for Nwchem runs."""

import os
import shutil
import subprocess

from monty.io import zopen
from monty.shutil import gzip_dir

from custodian.custodian import Job

__author__ = "Shyue Ping Ong"
__version__ = "0.1"
__maintainer__ = "Shyue Ping Ong"
__email__ = "ongsp@ucsd.edu"
__status__ = "Beta"
__date__ = "5/20/13"


class NwchemJob(Job):
    """A basic Nwchem job."""

    def __init__(
        self,
        nwchem_cmd,
        input_file="mol.nw",
        output_file="mol.nwout",
        gzipped=False,
        backup=True,
        settings_override=None,
    ):
        """Initialize a basic NwChem job.

        Args:
            nwchem_cmd ([str]): Command to run Nwchem as a list of args. For
                example, ["nwchem"].
            input_file (str): Input file to run. Defaults to "mol.nw".
            output_file (str): Name of file to direct standard out to.
                Defaults to "mol.nwout".
            backup (bool): Whether to backup the initial input files. If True,
                the input files will be copied with a ".orig" appended.
                Defaults to True.
            gzipped (bool): Deprecated. Please use the Custodian class's
                gzipped_output option instead.
            settings_override ([dict]):
                An ansible style list of dict to override changes.
                #TODO: Not implemented yet.
        """
        self.nwchem_cmd = nwchem_cmd
        self.input_file = input_file
        self.output_file = output_file
        self.backup = backup
        self.gzipped = gzipped
        self.settings_override = settings_override

    def setup(self, directory="./"):
        """Performs backup if necessary."""
        if self.backup:
            shutil.copy(os.path.join(directory, self.input_file), os.path.join(directory, f"{self.input_file}.orig"))

    def run(self, directory="./"):
        """Performs actual nwchem run."""
        with zopen(self.output_file, "w") as fout:
            return subprocess.Popen([*self.nwchem_cmd, self.input_file], cwd=directory, stdout=fout)  # pylint: disable=R1732

    def postprocess(self, directory="./"):
        """Renaming or gzipping as needed."""
        if self.gzipped:
            gzip_dir(directory)



"""This module implements specific error handler for FEFF runs."""

import logging
import os
import re

from pymatgen.io.feff.sets import FEFFDictSet

from custodian.custodian import ErrorHandler
from custodian.feff.interpreter import FeffModder
from custodian.utils import backup

__author__ = "Chen Zheng"
__copyright__ = "Copyright 2012, The Materials Project"
__version__ = "0.1"
__maintainer__ = "Chen Zheng"
__email__ = "chz022@ucsd.edu"
__date__ = "Oct 18, 2017"

FEFF_BACKUP_FILES = [
    "ATOMS",
    "HEADER",
    "PARAMETERS",
    "POTENTIALS",
    "feff.inp",
    "*.cif",
    "pot.bin",
]

logger = logging.getLogger(__name__)


class UnconvergedErrorHandler(ErrorHandler):
    """Correct the unconverged error of FEFF's SCF calculation."""

    is_monitor = False

    def __init__(self, output_filename="log1.dat"):
        """Initialize the handler with the output file to check.

        Args:
            output_filename (str): Filename for the log1.dat file. log1.dat file
                contains the SCF calculation convergence information. Change this only
                if it is different from the default (unlikely).
        """
        self.output_filename = output_filename

    def check(self, directory="./"):
        """
        If the FEFF run does not converge, the check will return
        "TRUE".
        """
        return self._notconverge_check(directory)

    def _notconverge_check(self, directory):
        # Process the output file and get converge information
        not_converge_pattern = re.compile("Convergence not reached.*")
        converge_pattern = re.compile("Convergence reached.*")
        with open(os.path.join(directory, self.output_filename)) as file:
            for line in file:
                if len(not_converge_pattern.findall(line)) > 0:
                    return True

                if len(converge_pattern.findall(line)) > 0:
                    return False
        return None

    def correct(self, directory="./"):
        """Perform the corrections."""
        backup(FEFF_BACKUP_FILES, directory=directory)
        feff_input = FEFFDictSet.from_directory(directory)
        scf_values = feff_input.tags.get("SCF")
        nscmt = scf_values[2]
        ca = scf_values[3]
        nmix = scf_values[4]
        actions = []

        # Add RESTART card to PARAMETERS
        if "RESTART" not in feff_input.tags:
            actions += [{"dict": "PARAMETERS", "action": {"_set": {"RESTART": []}}}]

        if nscmt < 100 and ca == 0.2:
            scf_values[2] = 100
            scf_values[4] = 3  # Set nmix = 3
            actions += [{"dict": "PARAMETERS", "action": {"_set": {"SCF": scf_values}}}]
            FeffModder(directory=directory).apply_actions(actions)
            return {"errors": ["Non-converging job"], "actions": actions}

        if nscmt == 100 and nmix == 3 and ca > 0.01:
            # Reduce the convergence accelerator factor
            scf_values[3] = round(ca / 2, 2)
            actions += [{"dict": "PARAMETERS", "action": {"_set": {"SCF": scf_values}}}]
            FeffModder(directory=directory).apply_actions(actions)
            return {"errors": ["Non-converging job"], "actions": actions}

        if nmix == 3 and ca == 0.01:
            # Set ca = 0.05 and set nmix
            scf_values[3] = 0.05
            scf_values[4] = 5
            actions += [{"dict": "PARAMETERS", "action": {"_set": {"SCF": scf_values}}}]
            FeffModder(directory=directory).apply_actions(actions)
            return {"errors": ["Non-converging job"], "actions": actions}

        if nmix == 5 and ca == 0.05:
            # Set ca = 0.05 and set nmix
            scf_values[3] = 0.05
            scf_values[4] = 10
            actions += [{"dict": "PARAMETERS", "action": {"_set": {"SCF": scf_values}}}]
            FeffModder(directory=directory).apply_actions(actions)
            return {"errors": ["Non-converging job"], "actions": actions}

        if nmix == 10 and ca < 0.2:
            # loop through ca with nmix = 10
            scf_values[3] = round(ca * 2, 2)
            actions += [{"dict": "PARAMETERS", "action": {"_set": {"SCF": scf_values}}}]
            FeffModder(directory=directory).apply_actions(actions)
            return {"errors": ["Non-converging job"], "actions": actions}

        # Unfixable error. Just return None for actions.
        return {"errors": ["Non-converging job"], "actions": None}
